/// <reference types="node" />
import { ITermios, INative } from './interfaces';
export declare const native: INative;
/**
 * Class holding `struct termios` data.
 */
export declare class Termios implements ITermios {
    private _data;
    /** Getter/setter for input flags. */
    get c_iflag(): number;
    set c_iflag(value: number);
    /** Getter/setter for output flags. */
    get c_oflag(): number;
    set c_oflag(value: number);
    /** Getter/setter for control flags. */
    get c_cflag(): number;
    set c_cflag(value: number);
    /** Getter/setter for local flags. */
    get c_lflag(): number;
    set c_lflag(value: number);
    /** Buffer to access control code settings. */
    get c_cc(): Buffer;
    /**
     * Create new termios object.
     *
     * `from` can be a valid file descriptor (number),
     * another `Termios` object (copy constructor) or `null` (all data zeroed out).
     * Omitting `from` will try to load default values from *ttydefaults.h*
     * (not supported by all platforms).
     *
     * @param from Optional argument to pull termios settings from.
     */
    constructor(from?: number | ITermios | null);
    /** Load termios data from file descriptor `fd`. */
    loadFrom(fd: number): void;
    /**
     * Write termios data to file descriptor `fd`.
     *
     * `action` should be one of `native.ACTION`
     * (default: `TCSAFLUSH`).
     */
    writeTo(fd: number, action?: number): void;
    /** Return input channel baud rate setting as in `native.BAUD`. */
    getInputSpeed(): number;
    /** Return output channel baud rate setting as in `native.BAUD`. */
    getOutputSpeed(): number;
    /**
     * Set input channel baud rate.
     *
     * `speed` should be one of the baudrates in `native.BAUD`.
     */
    setInputSpeed(speed: number): void;
    /**
     * Set output channel baud rate.
     *
     * `speed` should be one of the baudrates in `native.BAUD`.
     */
    setOutputSpeed(speed: number): void;
    /**
     * Set input and output channel baud rate.
     *
     * `speed` should be one of the baudrates in `native.BAUD`.
     * @note Normally the input and output speed are entangled by the system,
     * thus setting one would also change the other one.
     */
    setSpeed(speed: number): void;
    /** Convenient method to set termios data to raw mode (flags taken from Python). */
    setraw(): void;
    /** Convenient method to set termios data to cbreak mode (flags taken from Python). */
    setcbreak(): void;
    /**
     * Convenient method to set termios data back to cooked mode.
     *
     * @note This method enables typical flags of cooked mode,
     * like job control (BRKINT), NL/CR rewrites and the line editor
     * (ICANON) with echoing (ECHO). This might differ from your
     * expectations, as systems may use slightly different settings.
     */
    setcooked(): void;
}
